/*
 *  TU Eindhoven
 *  Eindhoven, The Netherlands
 *
 *  Name            :   latency_minimal.cc
 *
 *  Author          :   Sander Stuijk (sander@ics.ele.tue.nl)
 *
 *  Date            :   September 14, 2006
 *
 *  Function        :   Compute minimal latency of SDFG for all possible
 *                      executions.
 *
 *  History         :
 *      14-09-06    :   Initial version.
 *
 * $Id: latency_minimal.cc,v 1.1.1.1 2007/10/02 10:59:45 sander Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 * In other words, you are welcome to use, share and improve this program.
 * You are forbidden to forbid anyone else to use, share and improve
 * what you give them.   Happy coding!
 */

#include <math.h>  
#include <iostream>  
#include <assert.h>  
#include <time.h>  
#include <sys/resource.h>  

#include "statespace.h"

#define TDTIME_MAX INT_MAX

#define max(a,b) ((a)>(b) ? (a) : (b))
#define min(a,b) ((a)<(b) ? (a) : (b))

typedef unsigned short TCnt;
typedef unsigned long TBufSize;
typedef double TDtime;

static TimedSDFgraph *g;
static SDFactor *outputActor;
static SDFtime SDF_MAX_TIME;
static TCnt SDF_OUT_ACTOR_REP_CNT;

static
SDFactor *findOutputActor(TimedSDFgraph *g, RepetitionVector repVec)
{
    int min = INT_MAX;
    SDFactor *a = NULL;
    
    for (SDFactorsIter iter = g->actorsBegin(); iter != g->actorsEnd(); iter++)
    {
        if (repVec[(*iter)->getId()] < min)
        {
            a = *iter;
            min = repVec[a->getId()];
        }
    }
    
    return a;
}

static
SDFtime getMaxExecTime(TimedSDFgraph *g)
{
    SDFtime max = 0;
    
    for (SDFactorsIter iter = g->actorsBegin(); iter != g->actorsEnd(); iter++)
    {
        TimedSDFactor *a = (TimedSDFactor*)(*iter);
        
        if (a->getExecutionTime() > max)
            max = a->getExecutionTime();
    }
    
    return max;
}

/******************************************************************************   * State  
 *****************************************************************************/
typedef struct _State  
{  
  TCnt **act_clk;
  TDtime glb_clk;  
  TBufSize *ch; 
} State;  

/**  
 * clearState ()  
 * The function sets the state to zero.  
 */  
static
void clearState(State &s)  
{
    for (uint i = 0; i < g->nrChannels(); i++)
        s.ch[i] = 0;
    
    s.glb_clk = 0;

    for (uint i = 0; i < g->nrActors(); i++)
        for (TCnt t = 0; t < SDF_MAX_TIME+1; t++)
            s.act_clk[i][t] = 0;
}  

/**  
 * equalStates ()  
 * The function compares to states and returns true if they are equal.  
 */  
static inline  
bool equalStates(const State &s1, const State &s2)  
{
    if (s1.glb_clk != s2.glb_clk)
        return false;
    
    for (uint i = 0; i < g->nrChannels(); i++)
        if (s1.ch[i] != s2.ch[i])
            return false;
    
    for (uint i = 0; i < g->nrActors(); i++)
        for (uint t = 0; t < SDF_MAX_TIME+1; t++)
            if (s1.act_clk[i][t] != s2.act_clk[i][t])
                return false;

    return true;
}  

/**  
 * copyState ()  
 * The function copies the state.  
 */  
static inline  
void copyState(State &to, const State &from)  
{  
    to.glb_clk = from.glb_clk;
    
    for (uint i = 0; i < g->nrChannels(); i++)
        to.ch[i] = from.ch[i];
    
    for (uint i = 0; i < g->nrActors(); i++)
        for (uint t = 0; t < SDF_MAX_TIME+1; t++)
            to.act_clk[i][t] = from.act_clk[i][t];
}

/**  
 * createState ()  
 * The function allocates memory for a new state.
 */  
static
void createState(State &s)  
{
    s.ch = new TBufSize [g->nrChannels()];
    
    s.act_clk = new TCnt* [g->nrActors()];
    for (uint i = 0; i < g->nrActors(); i++)
        s.act_clk[i] = new TCnt [SDF_MAX_TIME+1];
}  

/**  
 * destroyState ()  
 * The function deallocates all data structures of the state
 */  
static
void destroyState(State &s)  
{
    delete [] s.ch;
    
    for (uint i = 0; i < g->nrActors(); i++)
        delete [] s.act_clk[i];
    delete [] s.act_clk;
}  

#if 0
/**
 * printState ()
 * The function outputs the state to the output stream.
 */
static
void printState(ostream &out, const State &s)  
{
    out << "((";
    for (uint i = 0; i < g->nrChannels(); i++)
    {
        out << s.ch[i];
        if (i != g->nrChannels() -1 )
            out << ", ";
    }
    out << "), " << s.glb_clk << ", (";
    for (uint i = 0; i < g->nrActors(); i++)
    {
        out << "{";
        for (TCnt t = 0; t < SDF_MAX_TIME+1; t++)
        {
            out << s.act_clk[i][t];
            if (t != SDF_MAX_TIME)
                out << ", ";
        }
        out << "}";
        if (i != g->nrActors() - 1)
            out << ", ";
    }
    out << "))" << endl;
}  
#endif

/******************************************************************************
 * Stack
 *****************************************************************************/

/**  
 * StackPos  
 * Index into stack.  
 */  
typedef long long StackPos;  

/**  
 * stack  
 * The stack...  
 */  
static State *stack;  

/**  
 * stackPtr  
 * Index into stack of the first free space.  
 */  
static StackPos stackPtr;  
static StackPos maxStackPtr;  
static StackPos STACK_SIZE;  

/**
 * resizeStack ()
 * Enlarge the stack with a factor of 2.
 */
static
void resizeStack()
{
    StackPos newStackSize = 2 * STACK_SIZE;
    State *newStack;  

    // Allocate new stack
    newStack = new State [newStackSize];
    
    // Copy old stack to new stack
    for (StackPos p = 0; p < STACK_SIZE; p++)
        newStack[p] = stack[p];
    
    // Create state for new part of stack
    for (StackPos p = STACK_SIZE; p < newStackSize; p++)
        createState(newStack[p]);
    
    // Cleanup old stack
    delete [] stack;
    
    // Put new stack in place
    stack = newStack;
    STACK_SIZE = newStackSize;
}

/**  
 * createStack ()  
 * The function creates a stack.  
 */  
static
void createStack()  
{  
    stack = new State [STACK_SIZE];

    if (stack == NULL)
        throw CException("Failed creating stack.");  

    for (uint i = 0; i < STACK_SIZE; i++)
        createState(stack[i]);

    maxStackPtr = 0;  
}

/**  
 * destroyStack()  
 * The function destroys a stack.  
 */
static  
void destroyStack()  
{  
    for (uint i = 0; i < STACK_SIZE; i++)
        destroyState(stack[i]);

    delete [] stack;
}  

/**  
 * clearStack ()  
 * Reset the stack to an empty state.  
 */
static
void clearStack()  
{  
    stackPtr = 0;     
}  

/**  
 * popStack ()  
 * Remove last element from the stack.  
 */  
static inline  
void popStack()  
{  
    if (stackPtr > 0)  
        stackPtr--;  
}  

/**  
 * pushStack ()  
 * Put an element on the stack.  
 */  
static inline  
void pushStack(const State &s)  
{  
    if (stackPtr >= STACK_SIZE)
    {
        cerr << "[INFO] buffy resizing stack." << endl;
        resizeStack();
    }
    
    copyState(stack[stackPtr], s);
    stackPtr++;

    if (stackPtr > maxStackPtr) maxStackPtr = stackPtr;
}  

/**  
 * topStack ()  
 * The function returns the last element of the stack.  
 */  
static inline  
State &topStack()  
{  
    return stack[stackPtr-1];  
}  
      /******************************************************************************
 * Hash
 *****************************************************************************/  
#define INVALID_HASH_KEY    NULL  
#define NO_SLOT_IN_HASH     -1  
        
/**  
 * HashKey  
 *  
 */  
typedef unsigned long long HashKey;  

/**  
 * HashSlot  
 *  
 */  
typedef struct _HashSlot  
{  
    StackPos value;  
    struct _HashSlot *next;  
} HashSlot;  

/**  
 * hashTable  
 * The hash table...  
 */  
static HashSlot **hashTable;  
static HashKey HASH_TABLE_SIZE;

/**  
 * createHashTable ()  
 * The function constructs a hash table.  
 */  
static
void createHashTable()  
{  
    hashTable = new HashSlot* [HASH_TABLE_SIZE];

    if (hashTable == NULL)
        throw CException("Failed creating hash table.");  

    for (uint i = 0; i < HASH_TABLE_SIZE; i++)  
        hashTable[i] = INVALID_HASH_KEY;  
}

/**  
 * destroyHashTable ()  
 * The function destoys a hash table.  
 */  
static
void destroyHashTable()  
{  
    HashSlot *s_cur, *s_next;  

    for (uint i = 0; i < HASH_TABLE_SIZE; i++)  
    {  
        for (s_cur = hashTable[i]; s_cur != INVALID_HASH_KEY; s_cur = s_next)  
        {  
            s_next = s_cur->next;  
            delete s_cur;  
        }  
    }  

    delete [] hashTable;  
}  
        
/**  
 * clearHashTable ()  
 * Resets the hash table to contain no keys.  
 */  
static
void clearHashTable()  
{  
    HashSlot *s_cur, *s_next;  

    for (uint i = 0; i < HASH_TABLE_SIZE; i++)  
    {  
        for (s_cur = hashTable[i]; s_cur != INVALID_HASH_KEY; s_cur = s_next)  
        {  
            s_next = s_cur->next;
            delete s_cur;  
        }  

        hashTable[i] = INVALID_HASH_KEY;  
    }  
}  
        
/**  
 * hash ()  
 * The hash function. It is a standard multiplication hash.  
 */  
static inline  
HashKey hash(const State &s)  
{  
    double key = 0;  
    long long key_part;  

    key_part = (long long)(s.glb_clk);
    key = key * 39164205.20662217 + double(key_part) * 0.6180339887;
    
    for (uint i = 0; i < g->nrActors(); i++)
    {
        for (TCnt t = 0; t < SDF_MAX_TIME+1; t++)
        {
            key_part = (long long)(s.act_clk[i][t]);
            key = key * 39164205.20662217 + double(key_part) * 0.6180339887;
        }
    }

    for (uint i = 0; i < g->nrChannels(); i++)
    {
        key_part = (long long)(s.ch[i]);
        key = key * 39164205.20662217 + double(key_part) * 0.6180339887;
    }

    key = fmod(key, 1);  
    return (HashKey) floor(HASH_TABLE_SIZE * key);  
}  
        
/**  
 * insertKeyHashTable ()  
 * The function inserts a (key, value) pair into the hash table.  
 */  
static inline  
void insertKeyHashTable(const HashKey key, const StackPos value)  
{  
    HashSlot *slot = new HashSlot;  
    slot->value = value;
    slot->next = hashTable[key];  

    // Insert the stack position in the hash table  
    hashTable[key] = slot;  
}  
        
/**  
 * findVallueInHashTable ()  
 * The function returns the StackPos at which the given (key,state) pair is  
 * located or NO_SLOT_IN_HASH if unkown.  
 */  
static inline  
StackPos findValueInHashTable(const HashKey key, const State &s)  
{  
    HashSlot *slot;  

    for (slot = hashTable[key]; slot != INVALID_HASH_KEY; slot = slot->next)  
    {  
        StackPos pos = slot->value;  

        if (equalStates(stack[pos], s))  
            return pos;  
    }  

    return NO_SLOT_IN_HASH;  
}  
      /******************************************************************************
 * SDF
 *****************************************************************************/      
/**  
 * outputIntervalCycle ()  
 * The function calculates the output interval of the states on the cycle. Its  
 * value is equal to the average time between two firings in the cycle.  
 */  
static inline  
TDtime outputIntervalCycle(const StackPos cyclePos)  
{  
    int nr_fire = 0;  
    TDtime time = 0;  

    // Check all state from stack till cycle complete  
    for (StackPos ptr = stackPtr -1; ptr >= cyclePos; ptr--)  
    {  
        // Output actor fired?  
        if (stack[ptr].act_clk[outputActor->getId()][0] > 0)  
            nr_fire++;  

        // Time between previous state  
        time += stack[ptr].glb_clk;  
    }  

    return time/(TDtime)(nr_fire);  
}  
        
/**  
 * storeState ()  
 * The function stores the given state at the stack and insert an entry  
 * into the hash table. Cycle detection is performed, as well as a  
 * check to verify wether a cycle is valid or not.  
 *  
 * Return values:  
 *      -1      - added state to stack  
 *      >= 0    - output interval of detected cycle  
 */  
static inline  
TDtime storeState(State &sdfState)  
{  
    // Calculate the key in the hash table  
    HashKey key = hash(sdfState);  

    // Check occurance in hash table  
    StackPos stackPos = findValueInHashTable(key, sdfState);  
    if (stackPos == NO_SLOT_IN_HASH)  
    {  
        // State not in hash table (i.e. state not seen before)  
        // Insert state in hash table  
        insertKeyHashTable(key, stackPtr);  

        // Insert state in stack  
        pushStack(sdfState);  
    }  
    else  
    {  
        // State has been visited before (i.e. found cycle)  
        return outputIntervalCycle(stackPos);  
    }  

    return -1;  
}  

#define GLB_CLK             sdfState.glb_clk  
#define ACT_CLK(a,t)        sdfState.act_clk[a][t]  
#define CH(c)               sdfState.ch[c]  

#define FIRE_ACT(a,t)       ACT_CLK(a,t)++;  
#define FIRE_ACT_END(a)     ACT_CLK(a,0)--;  
#define CH_TOKENS(c,n)      (CH(c) >= n)  
#define CONSUME(c,n)        CH(c) = CH(c) - n;
#define PRODUCE(c,n)        CH(c) = CH(c) + n;  
#define ACT_END_FIRE(a)     (ACT_CLK(a,0) != 0)  
#define ADVANCE_CLK         GLB_CLK = GLB_CLK + 1;  
#define NEXT_ITER           GLB_CLK = 0;  
#define LOWER_CLK(a)        for (TCnt t = 0; t < SDF_MAX_TIME; t++) \
                                ACT_CLK(a,t) = ACT_CLK(a,t+1); \
                            ACT_CLK(a,SDF_MAX_TIME) = 0;

static State sdfState;  
static State prevState;  

/**  
 * execSDFgraph()  
 * Execute the SDF graph till a deadlock is found or a recurrent state.  
 * The output interval (i.e. inverse of throughput) is returned.  
 */  
static
TDtime execSDFgraph(bool startFromInitialState = true)  
{   
    TDtime output_interval = -1;
    int repCnt = 0;  
    int globalTime = 0;
    
    if (startFromInitialState)
    {
        createState(sdfState);
        createState(prevState);  

        // Create initial state and add it to stack
        clearState(sdfState);  
        copyState(prevState, sdfState);  
        prevState.glb_clk = TDTIME_MAX;  

        // Initial tokens  
        for (SDFchannelsIter iter = g->channelsBegin(); 
                iter != g->channelsEnd(); iter++)
        {
            SDFchannel *ch = *iter;

            if (ch->getInitialTokens() != 0)
                PRODUCE(ch->getId(), ch->getInitialTokens());  
        }
    }
    
    // Fire the actors  
    while (true)  
    {  
        // Finish actor firings  
        for (SDFactorsIter iter = g->actorsBegin();
                iter != g->actorsEnd(); iter++)
        {
            SDFactor *a = *iter;
    
            while (ACT_END_FIRE(a->getId()))  
            {
                if (outputActor->getId() == a->getId())
                {
                    repCnt++;
                    if (repCnt == SDF_OUT_ACTOR_REP_CNT)  
                    {  
                        // Add state to hash of visited states  
                        output_interval = storeState(sdfState);  
                        NEXT_ITER;  
                        repCnt = 0;  
                    }  
                }

                for (SDFportsIter pIter = a->portsBegin();
                        pIter != a->portsEnd(); pIter++)
                {
                    SDFport *p = *pIter;
                    SDFchannel *ch = p->getChannel();

                    if (p->getType() == SDFport::Out)
                        PRODUCE(ch->getId(), p->getRate());  
                }

                FIRE_ACT_END(a->getId());  
            }  
        }

        // Start actor firings  
        for (SDFactorsIter iter = g->actorsBegin();
                iter != g->actorsEnd(); iter++)
        {
            TimedSDFactor *a = (TimedSDFactor*)*iter;

            while (true)
            {
                bool fire = true;
                
                // Check all input ports for tokens
                for (SDFportsIter pIter = a->portsBegin();
                        pIter != a->portsEnd(); pIter++)
                {
                    SDFport *p = *pIter;
                    SDFchannel *ch = p->getChannel();

                    if (p->getType() == SDFport::Out)
                        continue;

                    if (!CH_TOKENS(ch->getId(), p->getRate()))
                    {
                        fire = false;
                        break;
                    }
                }

                // Ready to fire?
                if (!fire)
                    break;

                // Consume tokens on each input  
                for (SDFportsIter pIter = a->portsBegin();
                        pIter != a->portsEnd(); pIter++)
                {
                    SDFport *p = *pIter;
                    SDFchannel *ch = p->getChannel();

                    if (p->getType() == SDFport::In)
                        CONSUME(ch->getId(), p->getRate());
                }

                // Fire the actor  
                FIRE_ACT(a->getId(), a->getExecutionTime());
            }
        }

        // Lower clocks  
        for (SDFactorsIter iter = g->actorsBegin();
                iter != g->actorsEnd(); iter++)
        {
            SDFactor *a = *iter;
            LOWER_CLK(a->getId());
        }

        // Advance the global time (this is total time since beginning 
        // of everything)
        globalTime++;  

        // Check for progress (i.e. no deadlock)  
        if (equalStates(prevState, sdfState))
        {
            return globalTime;
        }

        // Advance the global clock  
        ADVANCE_CLK;
        
        // Store state to check for progress  
        copyState(prevState, sdfState);  
    }
    
    destroyState(sdfState);
    destroyState(prevState);
}  

/**  
 * execSDFgraphUsingDemandList ()  
 * Execute the SDF graph in a self-timed manner. Only actors in the demand list
 * may be executed (upto the number of times specified). The function returns
 * the amount of time needed to execute all actors from the demand list. 
 * It starts from the current sdfState.
 */  
static
SDFtime execSDFgraphUsingDemandList(vector<uint> &demandList)  
{   
    TDtime output_interval = -1;
    SDFtime globalTime = 0;
    int repCnt = 0;  

    // Fire the actors  
    while (true)  
    {  
        // Finish actor firings  
        for (SDFactorsIter iter = g->actorsBegin();
                iter != g->actorsEnd(); iter++)
        {
            SDFactor *a = *iter;
    
            while (ACT_END_FIRE(a->getId()))  
            {
                if (outputActor->getId() == a->getId())
                {
                    repCnt++;
                    if (repCnt == SDF_OUT_ACTOR_REP_CNT)  
                    {  
                        // Add state to hash of visited states  
                        output_interval = storeState(sdfState);  
                        NEXT_ITER;  
                        repCnt = 0;  
                    }  
                }

                for (SDFportsIter pIter = a->portsBegin();
                        pIter != a->portsEnd(); pIter++)
                {
                    SDFport *p = *pIter;
                    SDFchannel *ch = p->getChannel();

                    if (p->getType() == SDFport::Out)
                        PRODUCE(ch->getId(), p->getRate());  
                }

                FIRE_ACT_END(a->getId());  
            }  
        }

        // Start actor firings  
        for (SDFactorsIter iter = g->actorsBegin();
                iter != g->actorsEnd(); iter++)
        {
            TimedSDFactor *a = (TimedSDFactor*)*iter;

            while (true)
            {
                bool fire = true;
                
                // Check all input ports for tokens
                for (SDFportsIter pIter = a->portsBegin();
                        pIter != a->portsEnd(); pIter++)
                {
                    SDFport *p = *pIter;
                    SDFchannel *ch = p->getChannel();

                    if (p->getType() == SDFport::Out)
                        continue;

                    if (!CH_TOKENS(ch->getId(), p->getRate()))
                    {
                        fire = false;
                        break;
                    }
                }

                // Ready to fire?
                if (!fire)
                    break;
                
                // Not allowed to fire according to demand list?
                if (demandList[a->getId()] == 0)
                    break;

                // Consume tokens on each input  
                for (SDFportsIter pIter = a->portsBegin();
                        pIter != a->portsEnd(); pIter++)
                {
                    SDFport *p = *pIter;
                    SDFchannel *ch = p->getChannel();

                    if (p->getType() == SDFport::In)
                        CONSUME(ch->getId(), p->getRate());
                }

                // Fire the actor  
                FIRE_ACT(a->getId(), a->getExecutionTime());
                
                // Decrease number of remaining actor firings in demand list
                demandList[a->getId()] = demandList[a->getId()] - 1;
            }
        }

        // Lower clocks  
        for (SDFactorsIter iter = g->actorsBegin();
                iter != g->actorsEnd(); iter++)
        {
            SDFactor *a = *iter;
            LOWER_CLK(a->getId());
        }

        // Check for progress (i.e. no deadlock)  
        if (equalStates(prevState, sdfState))
            return UINT_MAX;  

        // Advance the global clock  
        ADVANCE_CLK;

        // Store state to check for progress  
        copyState(prevState, sdfState);  
        
        // No actor firings left in demand list?
        bool noActorFiringsNeeded = true;
        for (SDFactorsIter iter = g->actorsBegin();
                iter != g->actorsEnd() && noActorFiringsNeeded; iter++)
        {
            TimedSDFactor *a = (TimedSDFactor*)*iter;
        
            if (demandList[a->getId()] != 0)
                noActorFiringsNeeded = false;
        }
        if (noActorFiringsNeeded)
            return globalTime;

        // Advance the global time (this is total time since beginning 
        // of everything)
        globalTime++;  
    }
    
    destroyState(sdfState);
    destroyState(prevState);
}  

/**
 * computeDemandList ()
 * The function computes the minimal number of firings required of each actor to  * enable the required number of firings of the destination actor connected to
 * the channel ch.
 */
static
void computeDemandList(const State &sdfState, const SDFchannel *ch,
        vector<uint> &demandList)
{
    SDFactor *srcActor, *dstActor;
    SDFport *srcPort, *dstPort;
    uint nrFiringsNeeded;
    int nrTokensNeeded;
    
    // Source and destination of channel
    srcActor = ch->getSrcActor();
    srcPort = ch->getSrcPort();
    dstActor = ch->getDstActor();
    dstPort = ch->getDstPort();
    
    // Number of tokens which need to be produced by source actor 
    // for destination actor
    nrTokensNeeded = dstPort->getRate() * demandList[dstActor->getId()];
    nrTokensNeeded = nrTokensNeeded - sdfState.ch[ch->getId()];
    
    // No tokens needed
    if (nrTokensNeeded < 0)
        return;
    
    // Compute number of firings needed of source actor
    nrFiringsNeeded = nrTokensNeeded / srcPort->getRate();
    if (nrTokensNeeded % srcPort->getRate() != 0)
        nrFiringsNeeded++;
    
    // Number of firings needed larger then what is currently known for source
    // actor?
    if (nrFiringsNeeded > demandList[srcActor->getId()])
    {
        // Update required number of firings of source actor
        demandList[srcActor->getId()] = nrFiringsNeeded;
        
        // Continue computing demand list from all channels incoming to source
        // actor
        for (SDFportsIter iter = srcActor->portsBegin();
                iter!= srcActor->portsEnd(); iter++)
        {
            SDFport *p = *iter;
            
            // Incoming channel?
            if (p->getType() == SDFport::In)
            {
                SDFchannel *c = p->getChannel();
                
                computeDemandList(sdfState, c, demandList);
            }
        }
    }
}
#include "sdf/output/xml/xml.h"
/**
 * stateSpaceLatencyAnalysisMinimal ()
 * Compute the latency of an SDF graph for unconstrained buffer sizes and
 * using auto-concurrency using a state-space traversal and self-timed 
 * execution.
 */
extern
double stateSpaceLatencyAnalysisMinimal(TimedSDFgraph *gr,
        SDFactor *srcActor, SDFactor *dstActor, double *thr,
        unsigned long long stackSz, unsigned long long hashSz)
{
    TimedSDFactor *srcDummyAct, *dstDummyAct;
    TimedSDFchannel *ch, *chDummyDst, *chDummySrc;
    vector<uint> demandList(gr->nrActors() + 2, 0);
    RepetitionVector repVec;
    double graphLatency;
    State deadSdfState, deadPrevState;
    
    // Copy arguments to globals
    STACK_SIZE = stackSz;
    HASH_TABLE_SIZE = hashSz;

    // Create a copy of the graph (the graph is modified in the function!)
    SDFcomponent comp(gr->getParent(), gr->getName(), gr->getId());
    g = gr->clone(comp);

    // Locate src and dst actor in copy
    srcActor = g->getActor(srcActor->getId());
    dstActor = g->getActor(dstActor->getId());
    
    // Check that the graph g is a strongly connected graph
    if (!isStronglyConnectedGraph(g))
        throw CException("Graph is not strongly connected.");
    
    // Find the output actor and its repetition count
    repVec = computeRepetitionVector(g);
    outputActor = findOutputActor(g, repVec);
    SDF_OUT_ACTOR_REP_CNT = repVec[outputActor->getId()];

    // Find largest execution time of all actors
    SDF_MAX_TIME = getMaxExecTime(g);

    // Add source and destination dummy actors
    srcDummyAct = g->createActor();
    dstDummyAct = g->createActor();

    TimedSDFactor::Processor proc;
    proc.type = "default";
    proc.execTime = 1;
    proc.stateSize = 0;

    srcDummyAct->addProcessor(&proc);
    srcDummyAct->setDefaultProcessor("default");
    dstDummyAct->addProcessor(&proc);
    dstDummyAct->setDefaultProcessor("default");

    // Connect dummy actors to real source and destination actors
    chDummySrc = g->createChannel(srcDummyAct, repVec[srcActor->getId()],
                                        srcActor, 1, 0);
    chDummyDst = g->createChannel(dstActor, 1, dstDummyAct,
                                        repVec[dstActor->getId()], 0);

    // Add self-loop to dummy source actor in graph g with no tokens
    ch = g->createChannel(srcDummyAct, 1, srcDummyAct, 1, 0);

    // Create hash and stack
    createStack(); clearStack();
    createHashTable(); clearHashTable();

    // Execute graph till deadlock
    execSDFgraph();

    // Compute also the throughput?
    if (thr != NULL)
    {
        // Store current and previous state 
        createState(deadSdfState);
        createState(deadPrevState);
        copyState(deadSdfState, sdfState);
        copyState(deadPrevState, prevState);
    }

    // Compute demand list
    demandList[dstDummyAct->getId()] = 1;
    computeDemandList(sdfState, chDummyDst, demandList);

    #ifdef VERBOSE
    for (SDFactorsIter iter = g->actorsBegin(); iter != g->actorsEnd(); iter++)
    {
        TimedSDFactor *a = (TimedSDFactor*)*iter;

        cerr << "demand[" << a->getId() << "] = ";
        cerr << demandList[a->getId()] << endl;
    }
    #endif
    
    // Fire the source actor once (i.e. put tokens on its output channel)
    sdfState.ch[chDummySrc->getId()] += repVec[srcActor->getId()];
    demandList[srcDummyAct->getId()] = 0;
    
    // Execute demand list in self-timed manner
    graphLatency = execSDFgraphUsingDemandList(demandList);

    // Compute also the throughput?
    if (thr != NULL)
    {
        // Restore the SDFG state to situation when deadlocked
        copyState(sdfState, deadSdfState);
        copyState(prevState, deadPrevState);
        destroyState(deadSdfState);
        destroyState(deadPrevState);

        // Fire the source actor once (i.e. put tokens on its output channel)
        sdfState.ch[chDummySrc->getId()] += repVec[srcActor->getId()];

        // Continue execution of the SDFG till a recurrent state is found
        // This will be the state in which the SDFG deadlocks (requiring the
        // source actor to fire again). The function returns the total amount
        // of time elapsed between the two deadlocks. In this time, the source
        // actor has fired once (i.e. throughput is inverse of the time
        // returned by the function).
        *thr = execSDFgraph(false);
        if (*thr != 0)
            *thr = 1.0 / *thr;
        else
            *thr = 0;
    }
    
    // Cleanup
    destroyStack();
    destroyHashTable();
    delete g;

    return graphLatency;
}

/**
 * stateSpaceLatencyAnalysisSingleProc ()
 * Compute the latency of an SDF graph for unconstrained buffer sizes and
 * using auto-concurrency using a state-space traversal and self-timed 
 * execution.
 */
extern
double stateSpaceLatencyAnalysisSingleProc(TimedSDFgraph *gr,
        SDFactor *srcActor, SDFactor *dstActor,
        unsigned long long stackSz, unsigned long long hashSz)
{
    TimedSDFactor *srcDummyAct, *dstDummyAct;
    TimedSDFchannel *ch, *chDummyDst, *chDummySrc;
    vector<uint> demandList(gr->nrActors() + 2, 0);
    RepetitionVector repVec;
    double graphLatency;
    
    // Copy arguments to globals
    g = gr;
    STACK_SIZE = stackSz;
    HASH_TABLE_SIZE = hashSz;

    // Check that the graph g is a strongly connected graph
    if (!isStronglyConnectedGraph(g))
        throw CException("Graph is not strongly connected.");
    
    // Find the output actor and its repetition count
    repVec = computeRepetitionVector(g);
    outputActor = findOutputActor(g, repVec);
    SDF_OUT_ACTOR_REP_CNT = repVec[outputActor->getId()];

    // Find largest execution time of all actors
    SDF_MAX_TIME = getMaxExecTime(g);

    // Add source and destination dummy actors
    srcDummyAct = g->createActor();
    dstDummyAct = g->createActor();

    TimedSDFactor::Processor proc;
    proc.type = "default";
    proc.execTime = 1;
    proc.stateSize = 0;

    srcDummyAct->addProcessor(&proc);
    srcDummyAct->setDefaultProcessor("default");
    dstDummyAct->addProcessor(&proc);
    dstDummyAct->setDefaultProcessor("default");

    // Connect dummy actors to real source and destination actors
    chDummySrc = gr->createChannel(srcDummyAct, repVec[srcActor->getId()],
                                        srcActor, 1, 0);
    chDummyDst = gr->createChannel(dstActor, 1, dstDummyAct,
                                        repVec[dstActor->getId()], 0);

    // Add self-loop to dummy source actor in graph g with no tokens
    ch = gr->createChannel(srcDummyAct, 1, srcDummyAct, 1, 0);

    // Create hash and stack
    createStack(); clearStack();
    createHashTable(); clearHashTable();
    
    // Execute graph till deadlock
    execSDFgraph();

    // Compute demand list
    demandList[dstDummyAct->getId()] = 1;
    computeDemandList(sdfState, chDummyDst, demandList);

    #ifdef VERBOSE
    for (SDFactorsIter iter = g->actorsBegin(); iter != g->actorsEnd(); iter++)
    {
        TimedSDFactor *a = (TimedSDFactor*)*iter;

        cerr << "demand[" << a->getId() << "] = ";
        cerr << demandList[a->getId()] << endl;
    }
    #endif
    
    // Fire the source actor once (i.e. put tokens on its output channel)
    sdfState.ch[chDummySrc->getId()] += repVec[srcActor->getId()];
    demandList[srcDummyAct->getId()] = 0;
    demandList[dstDummyAct->getId()] = 0;
    
    // Single processor latency is equal to the sum of the execution times of
    // all demanded actor firings
    graphLatency = 0;
    for (SDFactorsIter iter = g->actorsBegin(); iter!= g->actorsEnd(); iter++)
    {
        TimedSDFactor *a = (TimedSDFactor*)(*iter);
        
        graphLatency = graphLatency 
                            + demandList[a->getId()] * a->getExecutionTime();
    }
    
    // Cleanup
    destroyStack();
    destroyHashTable();
    
    return graphLatency;
}
