/*
 *  TU Eindhoven
 *  Eindhoven, The Netherlands
 *
 *  Name            :   latency_selftimed_minimal.cc
 *
 *  Author          :   Sander Stuijk (sander@ics.ele.tue.nl)
 *
 *  Date            :   November 8, 2006
 *
 *  Function        :   Compute latency of SDFG while using self-timed
 *                      execution both constraining the execution of the
 *                      source actor to achieve minimal latency with maximal
 *                      throughput.
 *
 *  History         :
 *      08-11-06    :   Initial version.
 *
 * $Id: latency_selftimed_minimal.cc,v 1.1.1.1 2007/10/02 10:59:45 sander Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 * In other words, you are welcome to use, share and improve this program.
 * You are forbidden to forbid anyone else to use, share and improve
 * what you give them.   Happy coding!
 */

#include <math.h>  
#include <iostream>  
#include <assert.h>  
#include <time.h>  
#include <sys/resource.h>  

#include "statespace.h"

#define TDTIME_MAX INT_MAX

#define max(a,b) ((a)>(b) ? (a) : (b))
#define min(a,b) ((a)<(b) ? (a) : (b))

typedef unsigned short TCnt;
typedef unsigned long TBufSize;
typedef double TDtime;

static TimedSDFgraph *g;
static SDFactor *outputActor;
static SDFtime SDF_MAX_TIME;
static TCnt SDF_OUT_ACTOR_REP_CNT;

static
SDFactor *findOutputActor(TimedSDFgraph *g, RepetitionVector repVec)
{
    int min = INT_MAX;
    SDFactor *a = NULL;
    
    for (SDFactorsIter iter = g->actorsBegin(); iter != g->actorsEnd(); iter++)
    {
        if (repVec[(*iter)->getId()] < min)
        {
            a = *iter;
            min = repVec[a->getId()];
        }
    }
    
    return a;
}

static
SDFtime getMaxExecTime(TimedSDFgraph *g)
{
    SDFtime max = 0;
    
    for (SDFactorsIter iter = g->actorsBegin(); iter != g->actorsEnd(); iter++)
    {
        TimedSDFactor *a = (TimedSDFactor*)(*iter);
        
        if (a->getExecutionTime() > max)
            max = a->getExecutionTime();
    }
    
    return max;
}

/******************************************************************************   * State  
 *****************************************************************************/
typedef struct _State  
{  
  TCnt **act_clk;
  TDtime glb_clk;  
  TBufSize *ch; 
} State;  

/**  
 * clearState ()  
 * The function sets the state to zero.  
 */  
static
void clearState(State &s)  
{
    for (uint i = 0; i < g->nrChannels(); i++)
        s.ch[i] = 0;
    
    s.glb_clk = 0;

    for (uint i = 0; i < g->nrActors(); i++)
        for (TCnt t = 0; t < SDF_MAX_TIME+1; t++)
            s.act_clk[i][t] = 0;
}  

/**  
 * equalStates ()  
 * The function compares to states and returns true if they are equal.  
 */  
static inline  
bool equalStates(const State &s1, const State &s2)  
{
    if (s1.glb_clk != s2.glb_clk)
        return false;
    
    for (uint i = 0; i < g->nrChannels(); i++)
        if (s1.ch[i] != s2.ch[i])
            return false;
    
    for (uint i = 0; i < g->nrActors(); i++)
        for (uint t = 0; t < SDF_MAX_TIME+1; t++)
            if (s1.act_clk[i][t] != s2.act_clk[i][t])
                return false;

    return true;
}  

/**  
 * copyState ()  
 * The function copies the state.  
 */  
static inline  
void copyState(State &to, const State &from)  
{  
    to.glb_clk = from.glb_clk;
    
    for (uint i = 0; i < g->nrChannels(); i++)
        to.ch[i] = from.ch[i];
    
    for (uint i = 0; i < g->nrActors(); i++)
        for (uint t = 0; t < SDF_MAX_TIME+1; t++)
            to.act_clk[i][t] = from.act_clk[i][t];
}

/**  
 * createState ()  
 * The function allocates memory for a new state.
 */  
static
void createState(State &s)  
{
    s.ch = new TBufSize [g->nrChannels()];
    
    s.act_clk = new TCnt* [g->nrActors()];
    for (uint i = 0; i < g->nrActors(); i++)
        s.act_clk[i] = new TCnt [SDF_MAX_TIME+1];
}  

/**  
 * destroyState ()  
 * The function deallocates all data structures of the state
 */  
static
void destroyState(State &s)  
{
    delete [] s.ch;
    
    for (uint i = 0; i < g->nrActors(); i++)
        delete [] s.act_clk[i];
    delete [] s.act_clk;
}  

/******************************************************************************
 * Stack
 *****************************************************************************/

/**  
 * StackPos  
 * Index into stack.  
 */  
typedef long long StackPos;  

/**  
 * stack  
 * The stack...  
 */  
static State *stack;  

/**  
 * stackPtr  
 * Index into stack of the first free space.  
 */  
static StackPos stackPtr;  
static StackPos maxStackPtr;  
static StackPos STACK_SIZE;  

/**
 * resizeStack ()
 * Enlarge the stack with a factor of 2.
 */
static
void resizeStack()
{
    StackPos newStackSize = 2 * STACK_SIZE;
    State *newStack;  

    // Allocate new stack
    newStack = new State [newStackSize];
    
    // Copy old stack to new stack
    for (StackPos p = 0; p < STACK_SIZE; p++)
        newStack[p] = stack[p];
    
    // Create state for new part of stack
    for (StackPos p = STACK_SIZE; p < newStackSize; p++)
        createState(newStack[p]);
    
    // Cleanup old stack
    delete [] stack;
    
    // Put new stack in place
    stack = newStack;
    STACK_SIZE = newStackSize;
}

/**  
 * createStack ()  
 * The function creates a stack.  
 */  
static
void createStack()  
{  
    stack = new State [STACK_SIZE];

    if (stack == NULL)
        throw CException("Failed creating stack.");  

    for (uint i = 0; i < STACK_SIZE; i++)
        createState(stack[i]);

    maxStackPtr = 0;  
}

/**  
 * destroyStack()  
 * The function destroys a stack.  
 */
static  
void destroyStack()  
{  
    for (uint i = 0; i < STACK_SIZE; i++)
        destroyState(stack[i]);

    delete [] stack;
}  

/**  
 * clearStack ()  
 * Reset the stack to an empty state.  
 */
static
void clearStack()  
{  
    stackPtr = 0;     
}  

/**  
 * popStack ()  
 * Remove last element from the stack.  
 */  
static inline  
void popStack()  
{  
    if (stackPtr > 0)  
        stackPtr--;  
}  

/**  
 * pushStack ()  
 * Put an element on the stack.  
 */  
static inline  
void pushStack(const State &s)  
{  
    if (stackPtr >= STACK_SIZE)
    {
        cerr << "[INFO] buffy resizing stack." << endl;
        resizeStack();
    }
    
    copyState(stack[stackPtr], s);
    stackPtr++;

    if (stackPtr > maxStackPtr) maxStackPtr = stackPtr;
}  

/**  
 * topStack ()  
 * The function returns the last element of the stack.  
 */  
static inline  
State &topStack()  
{  
    return stack[stackPtr-1];  
}  
      /******************************************************************************
 * Hash
 *****************************************************************************/  
#define INVALID_HASH_KEY    NULL  
#define NO_SLOT_IN_HASH     -1  
        
/**  
 * HashKey  
 *  
 */  
typedef unsigned long long HashKey;  

/**  
 * HashSlot  
 *  
 */  
typedef struct _HashSlot  
{  
    StackPos value;  
    struct _HashSlot *next;  
} HashSlot;  

/**  
 * hashTable  
 * The hash table...  
 */  
static HashSlot **hashTable;  
static HashKey HASH_TABLE_SIZE;

/**  
 * createHashTable ()  
 * The function constructs a hash table.  
 */  
static
void createHashTable()  
{  
    hashTable = new HashSlot* [HASH_TABLE_SIZE];

    if (hashTable == NULL)
        throw CException("Failed creating hash table.");  

    for (uint i = 0; i < HASH_TABLE_SIZE; i++)  
        hashTable[i] = INVALID_HASH_KEY;  
}

/**  
 * destroyHashTable ()  
 * The function destoys a hash table.  
 */  
static
void destroyHashTable()  
{  
    HashSlot *s_cur, *s_next;  

    for (uint i = 0; i < HASH_TABLE_SIZE; i++)  
    {  
        for (s_cur = hashTable[i]; s_cur != INVALID_HASH_KEY; s_cur = s_next)  
        {  
            s_next = s_cur->next;  
            delete s_cur;  
        }  
    }  

    delete [] hashTable;  
}  
        
/**  
 * clearHashTable ()  
 * Resets the hash table to contain no keys.  
 */  
static
void clearHashTable()  
{  
    HashSlot *s_cur, *s_next;  

    for (uint i = 0; i < HASH_TABLE_SIZE; i++)  
    {  
        for (s_cur = hashTable[i]; s_cur != INVALID_HASH_KEY; s_cur = s_next)  
        {  
            s_next = s_cur->next;
            delete s_cur;  
        }  

        hashTable[i] = INVALID_HASH_KEY;  
    }  
}  
        
/**  
 * hash ()  
 * The hash function. It is a standard multiplication hash.  
 */  
static inline  
HashKey hash(const State &s)  
{  
    double key = 0;  
    long long key_part;  

    key_part = (long long)(s.glb_clk);
    key = key * 39164205.20662217 + double(key_part) * 0.6180339887;
    
    for (uint i = 0; i < g->nrActors(); i++)
    {
        for (TCnt t = 0; t < SDF_MAX_TIME+1; t++)
        {
            key_part = (long long)(s.act_clk[i][t]);
            key = key * 39164205.20662217 + double(key_part) * 0.6180339887;
        }
    }

    for (uint i = 0; i < g->nrChannels(); i++)
    {
        key_part = (long long)(s.ch[i]);
        key = key * 39164205.20662217 + double(key_part) * 0.6180339887;
    }

    key = fmod(key, 1);  
    return (HashKey) floor(HASH_TABLE_SIZE * key);  
}  
        
/**  
 * insertKeyHashTable ()  
 * The function inserts a (key, value) pair into the hash table.  
 */  
static inline  
void insertKeyHashTable(const HashKey key, const StackPos value)  
{  
    HashSlot *slot = new HashSlot;  
    slot->value = value;
    slot->next = hashTable[key];  

    // Insert the stack position in the hash table  
    hashTable[key] = slot;  
}  
        
/**  
 * findVallueInHashTable ()  
 * The function returns the StackPos at which the given (key,state) pair is  
 * located or NO_SLOT_IN_HASH if unkown.  
 */  
static inline  
StackPos findValueInHashTable(const HashKey key, const State &s)  
{  
    HashSlot *slot;  

    for (slot = hashTable[key]; slot != INVALID_HASH_KEY; slot = slot->next)  
    {  
        StackPos pos = slot->value;  

        if (equalStates(stack[pos], s))  
            return pos;  
    }  

    return NO_SLOT_IN_HASH;  
}  
      /******************************************************************************
 * SDF
 *****************************************************************************/  
typedef struct _TimingConstraintFiring
{
    CId id;
    SDFtime timeTillNextFiring;
    uint nrFiringsEnabled;
    StackPos stackPos;
    bool visit;
    struct _TimingConstraintFiring *next;
} TimingConstraintFiring;
    
/**  
 * computeThroughput ()  
 * The function calculates the throughput of the states on the cycle. Its  
 * value is equal to the average number of firings of an actor per time unit.
 */  
static inline  
TDtime computeThroughput(const StackPos cyclePos)  
{  
	int nr_fire = 0;  
	TDtime time = 0;  

	// Check all state from stack till cycle complete  
	for (StackPos ptr = stackPtr -1; ptr >= cyclePos; ptr--)  
	{  
	    // Output actor fired?  
        if (stack[ptr].act_clk[outputActor->getId()][0] > 0)  
            nr_fire++;  

	    // Time between previous state  
	    time += stack[ptr].glb_clk;  
	}  

	return (TDtime)(nr_fire)/(time);  
}  
	    
/**  
 * storeState ()  
 * The function stores the given state at the stack and insert an entry  
 * into the hash table. Cycle detection is performed, as well as a  
 * check to verify wether a cycle is valid or not.  
 *  
 * Return values:  
 *      -1      - state did not exist on the stack; the state is added
 *      >= 0    - state did exist on the stack; position of state on stack is
 *                returned
 */  
static inline  
StackPos storeState(State &sdfState)  
{  
    // Calculate the key in the hash table  
    HashKey key = hash(sdfState);  

    // Check occurance in hash table  
    StackPos stackPos = findValueInHashTable(key, sdfState);  
    if (stackPos == NO_SLOT_IN_HASH)  
    {  
        // State not in hash table (i.e. state not seen before)  
        // Insert state in hash table  
        insertKeyHashTable(key, stackPtr);  

        // Insert state in stack  
        pushStack(sdfState);  
    }  
    else  
    {  
        // State has been visited before (i.e. found cycle)
        return stackPos;
    }  

    return -1;  
}  

#define GLB_CLK             sdfState.glb_clk  
#define ACT_CLK(a,t)        sdfState.act_clk[a][t]  
#define CH(c)               sdfState.ch[c]  

#define FIRE_ACT(a,t)       ACT_CLK(a,t)++;  
#define FIRE_ACT_END(a)     ACT_CLK(a,0)--;  
#define CH_TOKENS(c,n)      (CH(c) >= n)  
#define CONSUME(c,n)        CH(c) = CH(c) - n;
#define PRODUCE(c,n)        CH(c) = CH(c) + n;  
#define ACT_END_FIRE(a)     (ACT_CLK(a,0) != 0)  
#define ADVANCE_CLK         GLB_CLK = GLB_CLK + 1;  
#define NEXT_ITER           GLB_CLK = 0;  
#define LOWER_CLK(a)        for (TCnt t = 0; t < SDF_MAX_TIME; t++) \
                                ACT_CLK(a,t) = ACT_CLK(a,t+1); \
                            ACT_CLK(a,SDF_MAX_TIME) = 0;

static State sdfState;  
static State prevState;  

/**  
 * execSDFgraph()  
 * Execute the SDF graph till a deadlock is found or a recurrent state.  
 * The list of timestamps at which the destination fires is returned.  
 */  
static
TimingConstraintFiring *execSDFgraph(SDFactor *dstActor)  
{   
    TimingConstraintFiring *timeDstFire = NULL;
    TimingConstraintFiring *timeConstraint, *lastTimeConstraint = NULL;
    StackPos recurrentState = -1;
    SDFtime globalTime = 0;
    int repCnt = 0;  

    createState(sdfState);
    createState(prevState);  

    // Create initial state and add it to stack
    clearState(sdfState);  
    copyState(prevState, sdfState);  
    prevState.glb_clk = TDTIME_MAX;  

    // Initial tokens  
    for (SDFchannelsIter iter = g->channelsBegin(); 
            iter != g->channelsEnd(); iter++)
    {
        SDFchannel *ch = *iter;

        if (ch->getInitialTokens() != 0)
            PRODUCE(ch->getId(), ch->getInitialTokens());  
    }

    // Fire the actors  
    while (true)  
    {  
        // Finish actor firings  
        for (SDFactorsIter iter = g->actorsBegin();
                iter != g->actorsEnd(); iter++)
        {
            SDFactor *a = *iter;
    
            while (ACT_END_FIRE(a->getId()))  
            {
                if (outputActor->getId() == a->getId())
                {
                    repCnt++;
                    if (repCnt == SDF_OUT_ACTOR_REP_CNT)  
                    {
                        // Add state to hash of visited states  
                        recurrentState = storeState(sdfState);  
                        if (recurrentState != -1)
                        {
                            // Close cycle on firing times dst actor.
                            // Last timing constraint added is related to
                            // first timing constraint in the list which
                            // was added after the recurrent state
                            timeConstraint = timeDstFire;
                            while (timeConstraint != NULL)
                            {
                                if (timeConstraint->stackPos >= recurrentState)
                                {
                                    lastTimeConstraint->next = timeConstraint;
                                    return timeDstFire;
                                }

                                // Next
                                timeConstraint = timeConstraint->next;
                            }

                            throw CException("[ERROR] No recurrent state"
                                             " found.");
                        }
                        NEXT_ITER;  
                        repCnt = 0;  
                    }  
                }

                for (SDFportsIter pIter = a->portsBegin();
                        pIter != a->portsEnd(); pIter++)
                {
                    SDFport *p = *pIter;
                    SDFchannel *ch = p->getChannel();

                    if (p->getType() == SDFport::Out)
                        PRODUCE(ch->getId(), p->getRate());  
                }

                FIRE_ACT_END(a->getId());  
            }  
        }

        // Start actor firings  
        for (SDFactorsIter iter = g->actorsBegin();
                iter != g->actorsEnd(); iter++)
        {
            TimedSDFactor *a = (TimedSDFactor*)*iter;

            while (true)
            {
                bool fire = true;
                
                // Check all input ports for tokens
                for (SDFportsIter pIter = a->portsBegin();
                        pIter != a->portsEnd(); pIter++)
                {
                    SDFport *p = *pIter;
                    SDFchannel *ch = p->getChannel();

                    if (p->getType() == SDFport::Out)
                        continue;

                    if (!CH_TOKENS(ch->getId(), p->getRate()))
                    {
                        fire = false;
                        break;
                    }
                }

                // Ready to fire?
                if (!fire)
                    break;

                // Consume tokens on each input  
                for (SDFportsIter pIter = a->portsBegin();
                        pIter != a->portsEnd(); pIter++)
                {
                    SDFport *p = *pIter;
                    SDFchannel *ch = p->getChannel();

                    if (p->getType() == SDFport::In)
                        CONSUME(ch->getId(), p->getRate());
                }

                // Fire the actor
                FIRE_ACT(a->getId(), a->getExecutionTime());

                // Firing the destination actor?
                if (dstActor->getId() == a->getId())
                {
                    // Allocate memory for new timing constraint
                    timeConstraint = new TimingConstraintFiring;

                    // Global time indicates time between subsequent firings of
                    // dst actor
                    timeConstraint->timeTillNextFiring = globalTime;
                    globalTime = 0;
                    
                    // Remember last state stored (needed to close cycle)
                    timeConstraint->stackPos = stackPtr - 1;

                    // Default values                    
                    timeConstraint->nrFiringsEnabled = 0;
                    timeConstraint->visit = false;
                    timeConstraint->next = NULL;

                    // Add timing constraint to the end of the list
                    if (lastTimeConstraint == NULL)
                    {
                        timeConstraint->id = 0;
                        timeDstFire = timeConstraint;
                    }
                    else
                    {
                        timeConstraint->id = lastTimeConstraint->id + 1;
                        lastTimeConstraint->next = timeConstraint;
                    }
                    lastTimeConstraint = timeConstraint;
                }
            }
        }

        // Lower clocks  
        for (SDFactorsIter iter = g->actorsBegin();
                iter != g->actorsEnd(); iter++)
        {
            SDFactor *a = *iter;
            LOWER_CLK(a->getId());
        }

        // Check for progress (i.e. no deadlock)  
        if (equalStates(prevState, sdfState))
            return timeDstFire;  
        
        // Advance the global clock  
        ADVANCE_CLK;
        
        // Advance the global time (this is total time since beginning 
        // of everything)
        globalTime++;  

        // Store state to check for progress  
        copyState(prevState, sdfState);  
    }
    
    destroyState(sdfState);
    destroyState(prevState);
}  

/**  
 * execConstrainedSDFgraph()  
 * Execute the SDF graph till a deadlock is found or a recurrent state.  
 * The output interval (i.e. inverse of throughput) is returned. The graph is
 * executed in a self-timed manner. Except that the firings of the source actor
 * may only occur after the time at which the destination actor fires minus the
 * latency of the graph.
 */  
static
TDtime execConstrainedSDFgraph(SDFactor *srcActor,
        TimingConstraintFiring *timeSrcFire, SDFtime latency,
        SDFactor *dstActor, vector<SDFtime> &srcTimeFire, 
        vector<SDFtime> &dstTimeFire, uint distance)
{   
    long long globalTime = 0, previousStep = -((long long)(latency));
    uint nrAllowedSrcFirings = 0;
    StackPos recurrentState = -1;
    int repCnt = 0;  
    bool foundRecurrentState = false;
    
    createState(sdfState);
    createState(prevState);  

    // Create initial state and add it to stack
    clearState(sdfState);  
    copyState(prevState, sdfState);  
    prevState.glb_clk = TDTIME_MAX;  

    // Initial tokens  
    for (SDFchannelsIter iter = g->channelsBegin(); 
            iter != g->channelsEnd(); iter++)
    {
        SDFchannel *ch = *iter;

        if (ch->getInitialTokens() != 0)
            PRODUCE(ch->getId(), ch->getInitialTokens());  
    }

    // Enable all firings of source which can start from the beginning of time
    while (globalTime >= previousStep + timeSrcFire->timeTillNextFiring)
    {
        // Increase number of allowed firings
        nrAllowedSrcFirings += timeSrcFire->nrFiringsEnabled;

        // Is this an actual firings of the dummy source actor?
        if (timeSrcFire->nrFiringsEnabled > 0)
            srcTimeFire.push_back(globalTime);
        
        // Store global time at which current step occurs
        previousStep += timeSrcFire->timeTillNextFiring;

        // Next
        timeSrcFire = timeSrcFire->next;
    }

    // Fire the actors  
    while (true)  
    {  
        // Finish actor firings  
        for (SDFactorsIter iter = g->actorsBegin();
                iter != g->actorsEnd(); iter++)
        {
            SDFactor *a = *iter;
    
            while (ACT_END_FIRE(a->getId()))  
            {
                if (outputActor->getId() == a->getId())
                {
                    repCnt++;
                    if (repCnt == SDF_OUT_ACTOR_REP_CNT)  
                    {  
                        if (!foundRecurrentState)
                        {
                            // Add state to hash of visited states  
                            recurrentState = storeState(sdfState);  
                            if (recurrentState != -1)
                            {
                                foundRecurrentState = true;
                                //return computeThroughput(recurrentState);
                            }
                        }
                        NEXT_ITER;  
                        repCnt = 0;  
                    }  
                }

                for (SDFportsIter pIter = a->portsBegin();
                        pIter != a->portsEnd(); pIter++)
                {
                    SDFport *p = *pIter;
                    SDFchannel *ch = p->getChannel();

                    if (p->getType() == SDFport::Out)
                        PRODUCE(ch->getId(), p->getRate());  
                }

                FIRE_ACT_END(a->getId());  
            }  
        }

        // Start actor firings  
        for (SDFactorsIter iter = g->actorsBegin();
                iter != g->actorsEnd(); iter++)
        {
            TimedSDFactor *a = (TimedSDFactor*)*iter;

            while (true)
            {
                bool fire = true;
                
                // Check all input ports for tokens
                for (SDFportsIter pIter = a->portsBegin();
                        pIter != a->portsEnd(); pIter++)
                {
                    SDFport *p = *pIter;
                    SDFchannel *ch = p->getChannel();

                    if (p->getType() == SDFport::Out)
                        continue;

                    if (!CH_TOKENS(ch->getId(), p->getRate()))
                    {
                        fire = false;
                        break;
                    }
                }

                // Ready to fire?
                if (!fire)
                    break;

                // Source actor enabled?
                if (srcActor->getId() == a->getId())
                {
                    // Firings of actor allowed?
                    if (nrAllowedSrcFirings == 0)
                        break;
                    else
                        nrAllowedSrcFirings--;
                }

                // Consume tokens on each input  
                for (SDFportsIter pIter = a->portsBegin();
                        pIter != a->portsEnd(); pIter++)
                {
                    SDFport *p = *pIter;
                    SDFchannel *ch = p->getChannel();

                    if (p->getType() == SDFport::In)
                        CONSUME(ch->getId(), p->getRate());
                }

                // Fire the actor  
                FIRE_ACT(a->getId(), a->getExecutionTime());

                // Firing the destination actor?
                if (dstActor->getId() == a->getId())
                {
                    dstTimeFire.push_back(globalTime);

                    // Explored full state-space and found all destination
                    // firing, related to the source firings in the space?
                    if (dstTimeFire.size() >= srcTimeFire.size() + distance
                            && foundRecurrentState)
                    {
                        return computeThroughput(recurrentState);
                    }
                }
            }
        }

        // Lower clocks  
        for (SDFactorsIter iter = g->actorsBegin();
                iter != g->actorsEnd(); iter++)
        {
            SDFactor *a = *iter;
            LOWER_CLK(a->getId());
        }

        // Check for progress (i.e. no deadlock)  
        //if (equalStates(prevState, sdfState))
        //    return 0;  
        
        // Advance the global clock  
        ADVANCE_CLK;
        
        // Advance the global time
        globalTime++;

        // Enable all firings of source which can start from this point in time
        while (globalTime >= previousStep + timeSrcFire->timeTillNextFiring)
        {
            // Increase number of allowed firings
            nrAllowedSrcFirings += timeSrcFire->nrFiringsEnabled;

            // Is this an actual firings of the dummy source actor?
            if (timeSrcFire->nrFiringsEnabled > 0 && !foundRecurrentState)
                srcTimeFire.push_back(globalTime);

            // Store global time at which current step occurs
            previousStep += timeSrcFire->timeTillNextFiring;
        
            // Next
            timeSrcFire = timeSrcFire->next;
        }

        // Store state to check for progress  
        copyState(prevState, sdfState);  
    }
    
    destroyState(sdfState);
    destroyState(prevState);
}  

/**
 * stateSpaceMinimalLatencySelfTimed ()
 * Compute the minimal latency and maximal throughput of an SDF graph for
 * unconstrained buffer sizes and using auto-concurrency using a state-space
 * traversal and self-timed execution. The source actor is not allowed to
 * execute in a self-timed manner. Its execution is constraint to minimize the
 * latency.
 */
extern
double stateSpaceMinimalLatencySelfTimed(TimedSDFgraph *gr,
        SDFactor *srcActor, SDFactor *dstActor, double *thr,
        unsigned long long stackSz, unsigned long long hashSz)
{
    RepetitionVector repVec;
    TimingConstraintFiring *timeDstFire, *dstFirings, *lastSrcFirings, *x;
    TimingConstraintFiring *timeSrcFire, *srcFirings;
    SDFstateSpaceThroughputAnalysis thrAnalysisAlgo;
    SDFtime graphLatency, actorLatency, latency;
    vector<SDFtime> srcTimeFire, dstTimeFire;
    TimedSDFchannel *chDummyDst, *ch;
    TimedSDFactor *dstDummyActor;
    uint nrFirings, distance;
    
    // Copy arguments to globals
    STACK_SIZE = stackSz;
    HASH_TABLE_SIZE = hashSz;

    // Create a copy of the graph (the graph is modified in the function!)
    SDFcomponent comp(gr->getParent(), gr->getName(), gr->getId());
    g = gr->clone(comp);

    // Locate src and dst actor in copy
    srcActor = g->getActor(srcActor->getId());
    dstActor = g->getActor(dstActor->getId());

    // Check that the graph g is a strongly connected graph
    if (!isStronglyConnectedGraph(g))
        throw CException("Graph is not strongly connected.");

    // Find the output actor and its repetition count
    repVec = computeRepetitionVector(g);
    outputActor = findOutputActor(g, repVec);
    SDF_OUT_ACTOR_REP_CNT = repVec[outputActor->getId()];

    // Find largest execution time of all actors
    SDF_MAX_TIME = getMaxExecTime(g);

    // Add destination dummy actor
    dstDummyActor = g->createActor();

    TimedSDFactor::Processor proc;
    proc.type = "default";
    proc.execTime = 1;
    proc.stateSize = 0;

    dstDummyActor->addProcessor(&proc);
    dstDummyActor->setDefaultProcessor("default");

    // Connect dummy destination actor to real destination actor
    chDummyDst = g->createChannel(dstActor, 1, dstDummyActor,
                                        repVec[dstActor->getId()], 0);

    // Add self-loop to source actor in graph g with no tokens
    ch = g->createChannel(srcActor, 1, srcActor, 1, 0);

    outputActor = dstDummyActor;
    SDF_OUT_ACTOR_REP_CNT = 1;

    // Create hash and stack
    createStack(); clearStack();
    createHashTable(); clearHashTable();
    
    // Execute graph
    timeDstFire = execSDFgraph(dstDummyActor);

    // Distance is equal to number of firings of the destination actor
    distance = 0;
    dstFirings = timeDstFire;
    while (dstFirings != NULL)
    {
        // Next
        distance++;
        dstFirings = dstFirings->next;
    }

    // Remove self-loop on source actor
    ch->getSrcActor()->removePort(ch->getSrcPort()->getName());
    ch->getDstActor()->removePort(ch->getDstPort()->getName());  
    g->removeChannel(ch->getName());

    // Clear hash and stack    
    clearStack();
    clearHashTable();

    // Compute minimal latency between source and destination actor
    latency = (SDFtime)stateSpaceLatencyAnalysisMinimal(gr, srcActor, dstActor);

    // Execute graph to find all moments in time at which the dst actor fires
    timeDstFire = execSDFgraph(dstDummyActor);

    nrFirings = 0;
    dstFirings = timeDstFire;
    lastSrcFirings = NULL;
    timeSrcFire = NULL;
    while (dstFirings != NULL)
    {
        if (dstFirings->visit)
        {
            // Find matching firing in source
            // Length of cycle in source and destination has same size!
            uint lengthCycle = 0;
            x = dstFirings;
            do
            {
                x->visit = false;
                lengthCycle++;
                x = x->next;
            } while (x->visit);
            
            x = timeSrcFire;
            while (x->id != lastSrcFirings->id - lengthCycle + 1)
                x = x->next;
            lastSrcFirings->next = x;
            break;
        }
        
        if (nrFirings >= distance)
            dstFirings->visit = true;

        // Create new timing constraint for source actor firings
        srcFirings = new TimingConstraintFiring ;
        srcFirings->timeTillNextFiring = dstFirings->timeTillNextFiring;
        srcFirings->visit = false;

        if (nrFirings < distance)
        {
            srcFirings->nrFiringsEnabled = 0;
        }
        else
        {
            srcFirings->nrFiringsEnabled = repVec[srcActor->getId()];
        }

        // Add timing constraint to list of constraints
        if (lastSrcFirings == NULL)
        {
            srcFirings->id = 0;
            timeSrcFire = srcFirings;
        }
        else
        {
            srcFirings->id = lastSrcFirings->id + 1;
            lastSrcFirings->next = srcFirings;
        }
        lastSrcFirings = srcFirings;

        // Next
        nrFirings++;
        dstFirings = dstFirings->next;
    }
    
    // Clear hash and stack    
    clearStack();
    clearHashTable();
   
    // Execute graph with given timing constraints to compute throughput
    *thr = execConstrainedSDFgraph(srcActor, timeSrcFire, latency,
                    dstDummyActor, srcTimeFire, dstTimeFire, distance);

    // Compute latency
    graphLatency = 0;
    for (uint i = 0; i < srcTimeFire.size() 
            && i + distance < dstTimeFire.size(); i++)
    {   
        actorLatency = dstTimeFire[i + distance] - srcTimeFire[i];

        if (actorLatency > graphLatency)
            graphLatency = actorLatency;

        #ifdef VERBOSE
        cerr << (i) << " <-> ";
        cerr << (i + distance) << "\t";
        cerr << dstTimeFire[i + distance] << " - ";
        cerr << srcTimeFire[i] << endl;
        #endif
    }
        
    if (graphLatency != latency)
    {
        cerr << "[ERROR] latency: " << graphLatency;
        cerr << " (optimal: " << latency << ")" << endl;
    }

    // Compute maximal throughput
    double graphThr = thrAnalysisAlgo.analyze(gr);
    if (graphThr != *thr)
    {
        cerr << "[ERROR] throughput: " << *thr;
        cerr << " (optimal: " << graphThr << ")" << endl;
    }
        
    // Cleanup
    destroyStack();
    destroyHashTable();
    delete g;

    return graphLatency;
}
