/*
 *  TU Eindhoven
 *  Eindhoven, The Netherlands
 *
 *  Name            :   memory.cc
 *
 *  Author          :   Sander Stuijk (sander@ics.ele.tue.nl)
 *
 *  Date            :   February 7, 2007
 *
 *  Function        :   SDFG mapping to MP-SoC
 *
 *  History         :
 *      07-02-07    :   Initial version.
 *
 * $Id: memory.cc,v 1.1.1.1 2007/10/02 10:59:48 sander Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 * In other words, you are welcome to use, share and improve this program.
 * You are forbidden to forbid anyone else to use, share and improve
 * what you give them.   Happy coding!
 */

#include "flow.h"

/**
 * modelNonLocalMemoryAccesses ()
 * Model non-local memory accesses in the application graph.
 */
void SDF3Flow::modelNonLocalMemoryAccesses()
{
    cerr << "[WARNING] model non-local memory access not implemented." << endl;
    
    // Advance to next state in the flow
    setNextStateOfFlow(FlowComputeStorageDist);    
}

/**
 * computeStorageDistributions ()
 * Compute trade-off space between storage space allocated to channels and
 * maximal throughput of the application graph.
 */
void SDF3Flow::computeStorageDistributions()
{
    SDFstateSpaceBufferAnalysis bufferAnalysisAlgo;
    StorageDistributionSet *distributionSet = NULL;
    StorageDistribution *distr = NULL;
    TimedSDFgraph *g;
    
    // Output current state of the flow
    logInfo("Compute storage-space / throughput trade-offs.");

    // Assume absense of auto-concurrency (i.e. add self-edge to every actor)
    g = getAppGraph()->clone();
    for (SDFactorsIter iter = g->actorsBegin(); iter != g->actorsEnd(); iter++)
    {
        SDFactor *a = *iter;
        g->createChannel(a, 1, a, 1, 1);
    }
    
    // Set processor with lowest execution time as the default processor
    for (SDFactorsIter iter = g->actorsBegin(); iter != g->actorsEnd(); iter++)
    {
        TimedSDFactor *a = (TimedSDFactor*)(*iter);
        TimedSDFactor::Processor *defaultProc = NULL;
        
        for (TimedSDFactor::ProcessorsIter iterP = a->processorsBegin();
                iterP != a->processorsEnd(); iterP++)
        {
            TimedSDFactor::Processor *p = *iterP;
            
            if (defaultProc == NULL || defaultProc->execTime > p->execTime)
                defaultProc = p;
        }
        
        a->setDefaultProcessor(defaultProc->type);
    }    
    
    // Compute trade-off space
    minStorageDistributions = bufferAnalysisAlgo.analyze(g);

    // Update storage-space size to ignore added self-edges. These channels have
    // space for 2 tokens.
    distributionSet = minStorageDistributions;
    while (distributionSet != NULL)
    {
        // Update size of the set
        distributionSet->sz = distributionSet->sz - 2 * g->nrActors();
        
        // Update all distributions in the set
        distr = distributionSet->distributions;
        while (distr != NULL)
        {
            // Update size of the distribution
            distr->sz = distributionSet->sz;

            // Next distribution
            distr = distr->next;
        }
        
        // Next set
        distributionSet = distributionSet->next;
    }
    
    // No storage distribution selected so far
    selectedStorageDistributionSet = NULL;
    selectedStorageDistribution = NULL;
    
    // Advance to next state in the flow
    setNextStateOfFlow(FlowSelectStorageDist);    
}
