/*
 *  TU Eindhoven
 *  Eindhoven, The Netherlands
 *
 *  Name            :   cost_scheduler.h
 *
 *  Author          :   Sander Stuijk (sander@ics.ele.tue.nl)
 *
 *  Date            :   January 6, 2006
 *
 *  Function        :   Communication scheduling
 *
 *  History         :
 *      06-01-06    :   Initial version.
 *
 * $Id: cost_scheduler.h,v 1.1.1.1 2007/10/02 10:59:48 sander Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 * In other words, you are welcome to use, share and improve this program.
 * You are forbidden to forbid anyone else to use, share and improve
 * what you give them.   Happy coding!
 */

#ifndef SDF_RESOURCE_ALLOCATION_NOC_ALLOCATION_COST_SCHEDULER_H_INCLUDED
#define SDF_RESOURCE_ALLOCATION_NOC_ALLOCATION_COST_SCHEDULER_H_INCLUDED

#include "message.h"
#include "schedulingentity.h"
#include "arch_graph.h"
#include "packet.h"

class CostScheduler
{
public:
    // Constructor
    CostScheduler() {};
    
    // Destructor
    virtual ~CostScheduler() {};

    // Schedule function
    virtual bool schedule() = 0;

    // Create a duplicate (settings only) of the scheduler
    virtual CostScheduler *createDuplicate() = 0;
    
    // Architecture graph
    void setArchGraph(InterconnectGraph *g) { archGraph = g; };
    InterconnectGraph *getArchGraph() const { return archGraph; };
    
    // Messages
    void setMessages(Messages &m) { messages = m;};
    MessagesIter messagesBegin() { return messages.begin(); };
    MessagesIter messagesEnd() { return messages.end(); };
    MessagesCIter messagesBegin() const { return messages.begin(); };
    MessagesCIter messagesEnd() const { return messages.end(); };
            
    // Scheduling entities
    SchedulingEntitiesIter schedulingEntitiesBegin()
            { return schedulingEntities.begin(); };
    SchedulingEntitiesIter schedulingEntitiesEnd()
            { return schedulingEntities.end(); };    
    uint nrSchedulingEntities() const { return schedulingEntities.size(); };

protected:
    // Routing
    void findAllRoutes(const Node *src, const Node *dst, 
            const CSize maxDetour, bool exact, Routes &routes);
    void findRoutes(const Node *src, const Node *dst, const CSize minLength,
            const CSize maxLength, Route &route, Routes &routes);
    CSize getLengthShortestPathBetweenNodes(const Node *src, const Node *dst);

    // Slots
    uint minFreeSlotsOnLink(Link *l, const TTime startTime,
            const TTime duration);
    bool findSlotsOnRoute(SchedulingEntity *e, SlotReservations &s);
    uint nrSlotsRequired(const TTime time, const CSize size, 
            const uint nrPacketsPerSlotTable);
    void findFreePackets(SchedulingEntity *e, SlotReservations &slotsRoute,
            Packets &packets);
    SlotReservations findFreeSlotsOnRoute(const Route *r, const TTime startTime,
            const TTime duation);
    bool findSlotsOnRoute(SchedulingEntity *e,
            SlotReservations &slotsForPackets, SlotReservations &s);
    SlotReservations findFreeSlotsOnLink(const Link *l, const TTime startTime,
            const TTime duration);
    SlotReservations findFreeSlotsOnFirstLinkRoute(const Route *r, 
            const TTime startTime, const TTime duration);
    void raisePreferenceLevelSlotsOnRoute(SchedulingEntity *e);
    void lowerPreferenceLevelSlotsOnRoute(SchedulingEntity *e);

    // Stream
    void ripupStream(MessagesIter iterMsg);
    void findSlotsAllocatedForStream(SchedulingEntity *e,
            SlotReservations &slotsAllocated);
    void findFreeSlotsForStream(Route *r,
            SlotReservations &slotsRoute);

    // Conflicts
    double severityConflict(SchedulingEntity *e, Message *m, Route *r);
    void ripupScheduleEntity(MessagesIter iterMsg);

    // Cost
    void sortMessagesOnCost();
    void sortRoutesOnCost(Routes &routes, const TTime startTime,
            const TTime duration);
    
    // Scheduling entities
    bool findScheduleEntityForMessage(Message *m, const CSize maxDetour);
    TTime earliestStartTime(const Message *m) const;
    TTime maximalDuration(const Message *m, TTime startTime,
            CSize lengthRoute) const;
    TTime minimalDuration(const Message *m, TTime startTime,
            SlotReservations slotReservations) const;
    SchedulingEntity *getFirstScheduledMessageInStream(Message *m); 

    // Resource management
    void reserveResources(SchedulingEntity *e);
    void releaseResources(SchedulingEntity *e);

    // Routes (hash)
    void createRoutesHash(const CSize maxDetour);
    void destroyRoutesHash();
    
    // Print
    ostream &print(ostream &out) const;

protected:
    // Messages
    Messages messages;
    
private:
    // Architecture graph
    InterconnectGraph *archGraph;
    
    // Scheduling entities
    SchedulingEntities schedulingEntities;
    
    // Routes (hash)
    Routes ***routesHash;
    CSize **routesMinLengthHash;
};

#endif
