/*
 *  TU Eindhoven
 *  Eindhoven, The Netherlands
 *
 *  Name            :   problem.h
 *
 *  Author          :   Sander Stuijk (sander@ics.ele.tue.nl)
 *
 *  Date            :   October 4, 2006
 *
 *  Function        :   Schedule problem
 *
 *  History         :
 *      04-10-06    :   Initial version.
 *
 * $Id: problem.h,v 1.1.1.1 2007/10/02 10:59:47 sander Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 * In other words, you are welcome to use, share and improve this program.
 * You are forbidden to forbid anyone else to use, share and improve
 * what you give them.   Happy coding!
 */

#ifndef SDF_RESOURCE_ALLOCATION_NOC_ALLOCATION_PROBLEM_H_INCLUDED
#define SDF_RESOURCE_ALLOCATION_NOC_ALLOCATION_PROBLEM_H_INCLUDED

#include "cost_scheduler.h"

// Forward class definition
class ScheduleProblem;

/**
 * ScheduleSwitch
 * This data structure describes the time-relation between two scheduling
 * problems when switching from one to another one. The overlap is defined
 * as the amount of time start of the second schedule (to) overlaps with the
 * first schedule (from). When the overlap is equal or larger then the schedule
 * period of the 'from' or 'to' schedule, the two schedules overlap completly.
 */
typedef struct _ScheduleSwitch
{
    ScheduleProblem *from;
    ScheduleProblem *to;
    TTime overlap;
} ScheduleSwitch;

typedef list<ScheduleSwitch>                ScheduleSwitches;
typedef ScheduleSwitches::iterator          ScheduleSwitchesIter;
typedef ScheduleSwitches::const_iterator    ScheduleSwitchesCIter;

/**
 * ScheduleProblem
 * A scheduling problem (single set of messages)
 */
class ScheduleProblem
{
public:
    // Constructor
    ScheduleProblem(CString name, CNode *messagesNode, CNode *archGraphNode,
        CNode *systemUsageNode);
    
    // Destructor
    ~ScheduleProblem();

    // Solve single problem
    bool solve();
    bool isProblemSolved() { return problemSolvedFlag; };
    
    // Name
    CString getName() const { return scheduleName; };

    // Architecture graph
    InterconnectGraph *getArchGraph() const { return archGraph; };

    // Slot preferences
    void setPreferedSlots(InterconnectGraph *g);

    // Scheduler
    CostScheduler *getScheduler() const { return scheduler; };
    void setScheduler(CostScheduler *s);
    
    // Schedule switches
    void addScheduleSwitch(ScheduleSwitch &s){ scheduleSwitches.push_back(s); };

    // Schedule constraints (this adds all constraints which are given by the
    // this schedule on schedule p to the schedule p).
    void constrainScheduleProblem(ScheduleProblem *p);

private:
    // Name of the schedule
    CString scheduleName;
    
    // Period after which schedule repeats
    TTime schedulePeriod;    
    
    // Set of messages which must be scheduled
    Messages messages;

    // Architecture graph
    InterconnectGraph *archGraph;

    // The scheduler
    CostScheduler *scheduler;
    
    // Schedule switches
    ScheduleSwitches scheduleSwitches;
    
    // Flags
    bool problemSolvedFlag;
};

typedef list<ScheduleProblem*>              ScheduleProblems;
typedef ScheduleProblems::iterator          ScheduleProblemsIter;
typedef ScheduleProblems::const_iterator    ScheduleProblemsCIter;

/**
 * solveSetOfScheduleProblems ()
 * The function tries to find a valid schedule for each of the given
 * schedule problems. It considers relations between the problems. On success,
 * it return true. Else it returns false.
 */
bool solveSetOfScheduleProblems(ScheduleProblems &problems);

/**
 * createNetworkMappingNode ()
 * Create an XML node which describes the mapping of messages send over
 * the interconnect to the resources in the interconnect.
 */
CNode *createNetworkMappingNode(ScheduleProblems &problems);

/**
 * createNetworkUsageNode ()
 * Create an XML node which describes the usage of the links in the
 * interconnect.
 */
CNode *createNetworkUsageNode(ScheduleProblems &problems);

/**
 * outputNetworkBindingAsXML ()
 * Output the binding of messages to the interconnect as XML.
 */
void outputNetworkBindingAsXML(ScheduleProblems &problems, ostream &out);

/**
 * outputNetworkUsageAsXML ()
 * Output the usage of the interconnect as XML.
 */
void outputNetworkUsageAsXML(ScheduleProblems &problems, ostream &out);

#endif
